@@ -290,7 +290,7 @@ def is_last_li(li, meta_data, current_numId):
 
 
 @ensure_tag(['p'])
-def get_li_nodes(li, meta_data):
+def get_single_list_nodes_data(li, meta_data):
     """
     Find consecutive li tags that have content that have the same list id.
     """
@@ -559,7 +559,7 @@ def is_title(p):
 
 
 @ensure_tag(['r'])
-def get_raw_data(r):
+def get_text_run_content_data(r):
     """
     It turns out that r tags can contain both t tags and drawing tags. Since we
     need both, this function will return them in the order in which they are
@@ -905,7 +905,7 @@ def get_ordered_list_type(meta_data, numId, ilvl):
     return meta_data.numbering_dict[numId][ilvl]
 
 
-def get_list_data(li_nodes, meta_data):
+def build_list(li_nodes, meta_data):
     """
     Build the list structure and return the root list
     """
@@ -933,10 +933,10 @@ def get_list_data(li_nodes, meta_data):
     def _build_non_li_content(el, meta_data):
         w_namespace = get_namespace(el, 'w')
         if el.tag == '%stbl' % w_namespace:
-            new_el, visited_nodes = get_table_data(el, meta_data)
+            new_el, visited_nodes = build_table(el, meta_data)
             return etree.tostring(new_el), visited_nodes
         elif el.tag == '%sp' % w_namespace:
-            return get_p_data(el, meta_data), [el]
+            return get_element_content(el, meta_data), [el]
         if has_text(el):
             raise UnintendedTag('Did not expect %s' % el.tag)
 
@@ -974,7 +974,7 @@ def get_list_data(li_nodes, meta_data):
             list_contents = []
             current_ol.append(li_el)
         # Get the data needed to build the current list item
-        list_contents.append(get_p_data(
+        list_contents.append(get_element_content(
             li_node,
             meta_data,
         ))
@@ -1049,7 +1049,7 @@ def get_list_data(li_nodes, meta_data):
 
 
 @ensure_tag(['tr'])
-def get_tr_data(tr, meta_data, row_spans):
+def build_tr(tr, meta_data, row_spans):
     """
     This will return a single tr element, with all tds already populated.
     """
@@ -1084,15 +1084,18 @@ def get_tr_data(tr, meta_data, row_spans):
                 if is_li(td_content, meta_data):
                     # If it is a list, create the list and update
                     # visited_nodes.
-                    li_nodes = get_li_nodes(td_content, meta_data)
-                    list_el, list_visited_nodes = get_list_data(
+                    li_nodes = get_single_list_nodes_data(
+                        td_content,
+                        meta_data,
+                    )
+                    list_el, list_visited_nodes = build_list(
                         li_nodes,
                         meta_data,
                     )
                     visited_nodes.extend(list_visited_nodes)
                     texts.append(etree.tostring(list_el))
                 elif td_content.tag == '%stbl' % w_namespace:
-                    table_el, table_visited_nodes = get_table_data(
+                    table_el, table_visited_nodes = build_table(
                         td_content,
                         meta_data,
                     )
@@ -1103,7 +1106,7 @@ def get_tr_data(tr, meta_data, row_spans):
                     visited_nodes.append(td_content)
                     continue
                 else:
-                    text = get_p_data(
+                    text = get_element_content(
                         td_content,
                         meta_data,
                         is_td=True,
@@ -1131,7 +1134,7 @@ def get_tr_data(tr, meta_data, row_spans):
 
 
 @ensure_tag(['tbl'])
-def get_table_data(table, meta_data):
+def build_table(table, meta_data):
     """
     This returns a table object with all rows and cells correctly populated.
     """
@@ -1145,7 +1148,7 @@ def get_table_data(table, meta_data):
     for el in table:
         if el.tag == '%str' % w_namespace:
             # Create the tr element.
-            tr_el = get_tr_data(
+            tr_el = build_tr(
                 el,
                 meta_data,
                 row_spans,
@@ -1158,8 +1161,8 @@ def get_table_data(table, meta_data):
 
 
 @ensure_tag(['t'])
-def handle_t_tag(
-        t, parent, hyperlink_id, remove_bold, remove_italics, meta_data):
+def get_t_tag_content(
+        t, parent, remove_bold, remove_italics, meta_data):
     """
     Generate the string data that for this particular t tag.
     """
@@ -1170,12 +1173,7 @@ def handle_t_tag(
     # that is not valid XML.
     # cgi will replace things like & < > with &amp; &lt; &gt;
     text = cgi.escape(t.text)
-    if hyperlink_id is not None:
-        # The relationship_id is the href
-        if hyperlink_id in meta_data.relationship_dict:
-            href = meta_data.relationship_dict[hyperlink_id]
-            # Do not do any styling on hyperlinks
-            return '<a href="%s">%s</a>' % (href, text)
+
     # Wrap the text with any modifiers it might have (bold, italics or
     # underline)
     el_is_bold = not remove_bold and (
@@ -1195,8 +1193,89 @@ def _get_image_size_from_image(target):
     return image.size
 
 
-@ensure_tag(['p'])
-def get_p_data(p, meta_data, is_td=False):
+def build_hyperlink(el, meta_data):
+    # If we have a hyperlink we need to get relationship_id
+    r_namespace = get_namespace(el, 'r')
+    hyperlink_id = el.get('%sid' % r_namespace)
+
+    # Once we have the hyperlink_id then we need to replace the
+    # hyperlink tag with its child run tags.
+    content = get_element_content(
+        el,
+        meta_data,
+        remove_bold=True,
+        remove_italics=True,
+    )
+    if not content:
+        return ''
+    if hyperlink_id in meta_data.relationship_dict:
+        href = meta_data.relationship_dict[hyperlink_id]
+        # Do not do any styling on hyperlinks
+        return '<a href="%s">%s</a>' % (href, content)
+    return ''
+
+
+def build_image(el, meta_data):
+    image_id = get_image_id(el)
+    if image_id not in meta_data.relationship_dict:
+        # This image does not have an image_id
+        return ''
+    src = meta_data.image_handler(
+        image_id,
+        meta_data.relationship_dict,
+    )
+    if image_id in meta_data.image_sizes:
+        width, height = meta_data.image_sizes[image_id]
+    else:
+        target = meta_data.relationship_dict[image_id]
+        width, height = _get_image_size_from_image(target)
+    # Make sure the width and height are not zero
+    if all((width, height)):
+        return '<img src="%s" height="%d" width="%d" />' % (
+            src,
+            height,
+            width,
+        )
+    else:
+        return '<img src="%s" />' % src
+    return ''
+
+
+def get_text_run_content(el, meta_data, remove_bold, remove_italics):
+    w_namespace = get_namespace(el, 'w')
+    text_output = ''
+    for child in get_text_run_content_data(el):
+        if child.tag == '%st' % w_namespace:
+            text_output += get_t_tag_content(
+                child,
+                el,
+                remove_bold,
+                remove_italics,
+                meta_data,
+            )
+        elif child.tag == '%sbr' % w_namespace:
+            text_output += '<br />'
+        elif child.tag in (
+                '%spict' % w_namespace,
+                '%sdrawing' % w_namespace,
+        ):
+            text_output += build_image(child, meta_data)
+        else:
+            raise SyntaxNotSupported(
+                '"%s" is not a supported content-containing '
+                'text run child.' % child.tag
+            )
+    return text_output
+
+
+@ensure_tag(['p', 'ins', 'smartTag', 'hyperlink'])
+def get_element_content(
+        p,
+        meta_data,
+        is_td=False,
+        remove_italics=False,
+        remove_bold=False,
+):
     """
     P tags are made up of several runs (r tags) of text. This function takes a
     p tag and constructs the text that should be part of the p tag.
@@ -1204,8 +1283,6 @@ def get_p_data(p, meta_data, is_td=False):
     image_handler should be a callable that returns the desired ``src``
     attribute for a given image.
     """
-    remove_italics = False
-    remove_bold = False
 
     # Only remove bold or italics if this tag is an h tag.
     # Td elements have the same look and feel as p/h elements. Right now we are
@@ -1219,106 +1296,45 @@ def get_p_data(p, meta_data, is_td=False):
     w_namespace = get_namespace(p, 'w')
     if len(p) == 0:
         return ''
-    child = p[0]
-    tags = (
+    # Only these tags contain text that we care about (eg. We don't care about
+    # delete tags)
+    content_tags = (
         '%sr' % w_namespace,
         '%shyperlink' % w_namespace,
         '%sins' % w_namespace,
         '%ssmartTag' % w_namespace,
     )
-    elements = []
-    # Get the tags that are r tags or hyperlink tags
-    while True:
+    elements_with_content = []
+    for child in p:
         if child is None:
             break
-        if child.tag in tags:
-            # By default nothing needs to be forced bold.
-            elements.append(child)
-        child = child.getnext()
+        if child.tag in content_tags:
+            elements_with_content.append(child)
 
-    # Loop through each of the r and hyperlink tags
-    for el in elements:
-        hyperlink_id = None
+    # Gather the content from all of the children
+    for el in elements_with_content:
         # Hyperlinks and insert tags need to be handled differently than
-        # normals runs.
+        # r and smart tags.
         if el.tag in ('%sins' % w_namespace, '%ssmartTag' % w_namespace):
-            # Insert tags can have an arbitrary number of r tags in them. Find
-            # each and insert them into the elements list as the next elements
-            # in reverse order.
-            el_index = elements.index(el)
-            for r in reversed(el.xpath('.//w:r', namespaces=el.nsmap)):
-                # Be very careful when editing around this. This could cause
-                # problems if not understood. We are intentionally inserting
-                # new elements into the currently looping list.
-                elements.insert(el_index + 1, r)
-            continue
+            p_text += get_element_content(
+                el,
+                meta_data,
+                remove_bold=remove_bold,
+                remove_italics=remove_italics,
+            )
         elif el.tag == '%shyperlink' % w_namespace:
-            # If we have a hyperlink we need to get relationship_id
-            r_namespace = get_namespace(el, 'r')
-            hyperlink_id = el.get('%sid' % r_namespace)
-
-            # Once we have the hyperlink_id then we need to replace the
-            # hyperlink tag with its child run tags.
-            text = ''
-            r = None
-            for r in el.xpath('.//w:r', namespaces=el.nsmap):
-                for child in get_raw_data(r):
-                    text += handle_t_tag(child, r, None, True, True, meta_data)
-            if r is None:
-                if has_text(el):
-                    # If there is text in this hyperlink we need to raise an
-                    # exception so that we don't lose content.
-                    raise SyntaxNotSupported(
-                        'Hyperlink with text outside run tags not supported.',
-                    )
-                # It is very likely that this was a hyperlink tag that had its
-                # content removed, office does not do a very good job at
-                # cleaning up old tags, as such this tag has no content and
-                # should be ignored.
-                continue
-            else:
-                t_el = r.find('%st' % w_namespace)
-                if t_el is None:
-                    continue
-                t_el.text = text
-            el = r
-
-        # t tags hold all the text content.
-        for child in get_raw_data(el):
-            if child.tag == '%st' % w_namespace:
-                p_text += handle_t_tag(
-                    child,
-                    el,
-                    hyperlink_id,
-                    remove_bold,
-                    remove_italics,
-                    meta_data,
-                )
-            elif child.tag == '%sbr' % w_namespace:
-                p_text += '<br />'
-            else:  # We have an image
-                image_id = get_image_id(child)
-                if image_id not in meta_data.relationship_dict:
-                    # This image does not have an image_id
-                    continue
-                src = meta_data.image_handler(
-                    image_id,
-                    meta_data.relationship_dict,
-                )
-                if image_id in meta_data.image_sizes:
-                    width, height = meta_data.image_sizes[image_id]
-                else:
-                    target = meta_data.relationship_dict[image_id]
-                    width, height = _get_image_size_from_image(target)
-                # Make sure the width and height are not zero
-                if all((width, height)):
-                    p_text += '<img src="%s" height="%d" width="%d" />' % (
-                        src,
-                        height,
-                        width,
-                    )
-                else:
-                    p_text += '<img src="%s" />' % src
+            p_text += build_hyperlink(el, meta_data)
+        elif el.tag == '%sr' % w_namespace:
+            p_text += get_text_run_content(
+                el,
+                meta_data,
+                remove_bold=remove_bold,
+                remove_italics=remove_italics,
+            )
+        else:
+            raise SyntaxNotSupported(
+                'Content element "%s" not handled.' % el.tag
+            )
 
     # This function does not return a p tag since other tag types need this as
     # well (td, li).
@@ -1407,7 +1423,7 @@ def create_html(tree, meta_data):
             continue
         header_value = is_header(el, meta_data)
         if is_header(el, meta_data):
-            p_text = get_p_data(el, meta_data)
+            p_text = get_element_content(el, meta_data)
             if p_text == '':
                 continue
             new_html.append(
@@ -1423,15 +1439,15 @@ def create_html(tree, meta_data):
                 continue
             if is_li(el, meta_data):
                 # Parse out the needed info from the node.
-                li_nodes = get_li_nodes(el, meta_data)
-                new_el, list_visited_nodes = get_list_data(
+                li_nodes = get_single_list_nodes_data(el, meta_data)
+                new_el, list_visited_nodes = build_list(
                     li_nodes,
                     meta_data,
                 )
                 visited_nodes.extend(list_visited_nodes)
             # Handle generic p tag here.
             else:
-                p_text = get_p_data(el, meta_data)
+                p_text = get_element_content(el, meta_data)
                 # If there is not text do not add an empty tag.
                 if p_text == '':
                     continue
@@ -1440,7 +1456,7 @@ def create_html(tree, meta_data):
             new_html.append(new_el)
 
         elif el.tag == '%stbl' % w_namespace:
-            table_el, table_visited_nodes = get_table_data(
+            table_el, table_visited_nodes = build_table(
                 el,
                 meta_data,
             )
